/* oceano 
 * (c) 2004 Jim Ursetto
 *
 * A very basic specular highlighting example.
 */

#include <kos.h>
#include <math.h>
#include <stdlib.h> /* for randnum */
#include <png/png.h>
#include <plx/texture.h>
#include <plx/context.h>
#include <plx/prim.h>
#include "plx-spec.h"

extern uint8 romdisk[];
KOS_INIT_FLAGS(INIT_DEFAULT | INIT_MALLOCSTATS);
KOS_INIT_ROMDISK(romdisk);

#define ENABLE_SCREENSHOT  /* screenshot to /pc/tmp with B button */
#define BORDER_TIMING 0    /* overscan timing */

#define vid_border_timing(r,g,b) if (border_timing) { vid_border_color(r, g, b); }

typedef struct {
	int x, y;    /* not used anymore */
	uint32 offset;
} sparkle_t;

#define WIGGLE 60.0f /* wave wiggle; higher = less */
#define SPEED 300.0f /* wave speed; higher = slower */
#define wiggle_calc(radians) (fsin((radians) * 2 * M_PI) / WIGGLE + 1.0f)
#define sparkle_index(x,y) sparkle_idx[(y)*(xsquares+1)+(x)]
/* Number of sparkles; there should be little performance impact from
 * raising this number. */
#define SPARKLES 10
/* Resolution of mesh; x=1 y=20 is enough to make a convincing wave,
 * but higher geometry improves the specular highlights. */
const int xsquares = 135;  /* 135 is about the maximum */
const int ysquares = 135;

static sparkle_t sparkle[SPARKLES];
/* sparkle_idx maps a mesh point to a sparkle--it's the index into the
 * global sparkle[] array.  Results are 1-based (0 means non-existent)
 * so subtract one before using */
static uint8 *sparkle_idx;

void init_sparkles() {
	int i;
	sparkle_idx = calloc((xsquares + 1) * (ysquares + 1), 1);

	for (i=0;i<SPARKLES;i++) {
		sparkle[i].x = randnum(xsquares+1);
		sparkle[i].y = randnum(ysquares+1);
		sparkle_index(sparkle[i].x, sparkle[i].y) = i+1;
		if (i < SPARKLES / 2) {
			sparkle[i].offset = 0xff808040;  /* synchronized, but out of phase */
		} else {
			sparkle[i].offset = 0xffffff80;  /* so only half fade out at a time */
		}
	}
}

void draw_mesh() {
	static float r = 0.0f;
	int i;
	int idx;
	int xsq, ysq;
	float x, y, z, u, v, zold;
	float dx, dy, du, dv;
	uint32 color = 0xffffffff;
	uint32 offset = 0xffffffff;
	plx_dr_state_t dr;

	plx_mat_identity();
	plx_mat3d_apply(PLX_MAT_SCREENVIEW);
	plx_mat3d_apply(PLX_MAT_PROJECTION);
	plx_mat3d_apply(PLX_MAT_MODELVIEW);

	plx_dr_init(&dr);

	y = -1.0f;
	v = 1.0f;
	dx = 2.0f / xsquares; /* xrange: -1.0f to 1.0f */
	dy = 2.0f / ysquares; /* yrange: -1.0f to 1.0f */
	du = 1.0f / xsquares; /* urange:  0.0f to 1.0f */
	dv = 1.0f / ysquares; /* vrange:  1.0f to 0.0f */

	zold = wiggle_calc(v+r);

	for (i=0;i<SPARKLES;i++) {
		/* Check for dead sparkles */
		if (sparkle[i].offset <= 0xff000000) {
			sparkle_index(sparkle[i].x, sparkle[i].y) = 0; /* death */
			sparkle[i].x = randnum(xsquares+1);
			sparkle[i].y = randnum(ysquares+1);
			sparkle_index(sparkle[i].x, sparkle[i].y) = i+1; /* resurrection */
			sparkle[i].offset = 0xffffff80;
		}
	}

	for (ysq = 0; ysq <= ysquares; ysq++) {
		x = -1.0f;
		u = 0.0f;
		z = wiggle_calc(v-dv+r);
		for (xsq = 0; xsq <= xsquares; xsq++) {
			idx = sparkle_index(xsq, ysq+1);
			if (idx)
				offset = sparkle[idx-1].offset;
			else offset = 0;
			plx_vert_ifdm3s(&dr, PLX_VERT, x, y+dy, z, color, u, v-dv, offset);

			idx = sparkle_index(xsq, ysq);
			if (idx)
				offset = sparkle[idx-1].offset;
			else offset = 0;
			plx_vert_ifdm3s(&dr, xsq == xsquares ? PLX_VERT_EOS : PLX_VERT,
					            x, y, zold, color, u, v, offset);
			x += dx;
			u += du;
			offset = 0;
		}
		y += dy;
		v -= dv;
		zold = z;
	}
	r += M_PI / SPEED;
	for (i=0;i<SPARKLES;i++) 
		sparkle[i].offset -= 0x101008;
}

pvr_init_params_t pvr_params = {
    /* Enable opaque and translucent polygons with size 16 */
    { PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_0 },

    /* Vertex buffer size -- increased over the default 512K (uses 1.2M currently) */
    2048*1024
};

int main(int argc, char **argv) {
	plx_texture_t * txr;
	int done;
	int border_timing = BORDER_TIMING;
	pvr_stats_t stats;

	srand((int)time(NULL));

	// Init PVR
    if (pvr_init(&pvr_params) < 0)
        return -1;

	// Load a texture
	txr = plx_txr_load("/rd/water.png", 0, 0);
	if (!txr)
		return 0;

	// Setup a texture context
	plx_cxt_init();
	plx_cxt_texture(txr);
	plx_cxt_culling(PLX_CULL_NONE);
	plx_cxt_specular(PLX_SPECULAR); /* enable specular highlighting */
	plx_cxt_fog(PLX_FOG_TABLE);

	// Init 3D matrices
	plx_mat3d_init();
	plx_mat3d_mode(PLX_MAT_PROJECTION);
	plx_mat3d_identity();
	plx_mat3d_perspective(45.0f, 640.0f / 480.0f, 0.1f, 100.0f);
	plx_mat3d_mode(PLX_MAT_MODELVIEW);

	pvr_fog_table_color(1.0f, 0.2f, 0.6f, 1.0f);
	pvr_fog_table_linear(1.8f, 2.2f);   /* haze on the horizon */
	pvr_set_bg_color(0.2f, 0.6f, 1.0f); /* sky */

	init_sparkles();

	// Until the user hits start...
	for (done = 0; !done; ) {
		// Check for start
		MAPLE_FOREACH_BEGIN(MAPLE_FUNC_CONTROLLER, cont_state_t, st)
			if (st->buttons & CONT_START)
				done = 1;
#ifdef ENABLE_SCREENSHOT
			if (st->buttons & CONT_B) {
				vid_screen_shot("/pc/tmp/oceano.ppm");
				done = 1;
			}
#endif
		MAPLE_FOREACH_END()

		// Setup the frame
		vid_border_timing(0, 0, 255);
		pvr_wait_ready();
		pvr_scene_begin();
		pvr_list_begin(PVR_LIST_OP_POLY);
		vid_border_timing(0, 128, 255);

		// Submit the context
		plx_cxt_send(PVR_LIST_OP_POLY);

		plx_mat3d_identity();
		plx_mat3d_translate(0.0f, 0.0f, -5.0f);
		plx_mat3d_rotate(70, -1, 0, 0);
		plx_mat3d_translate(0.0f, -3.6f, 0.6f);
		plx_mat3d_scale(1.1f, 1.0f, 1.0f); /* so edges aren't on screen */
		draw_mesh();

		pvr_scene_finish();
	}

	pvr_get_stats(&stats);
	dbglog(DBG_DEBUG, "3D Stats: %ld vblanks, frame rate ~%f fps, max vertex used %d bytes\n",
			stats.vbl_count, (double)stats.frame_rate, stats.vtx_buffer_used_max);

	return 0;
}
